Switch to using a distance algorithm in similar().

--- Linux-PAM-0.75/modules/pam_cracklib/pam_cracklib.c
+++ Linux-PAM-0.75/modules/pam_cracklib/pam_cracklib.c
@@ -1,6 +1,7 @@
 /* pam_cracklib module */
 
 /*
+ * 0.9. switch to using a distance algorithm in similar()
  * 0.85.  added six new options to use this with long passwords.
  * 0.8. tidied output and improved D(()) usage for debugging.
  * 0.7. added support for more obscure checks for new passwd.
@@ -50,8 +51,14 @@
 #define CRACKLIB_DICTPATH "/usr/share/dict/cracklib_dict"
 #endif
 
-#define PROMPT1 "New %s password: "
-#define PROMPT2 "Retype new %s password: "
+#ifdef MIN
+#undef MIN
+#endif
+
+#define MIN(_a, _b) (((_a) < (_b)) ? (_a) : (_b))
+
+#define PROMPT1 "New %s%spassword: "
+#define PROMPT2 "Retype new %s%spassword: "
 #define MISTYPED_PASS "Sorry, passwords do not match"
 
 /*
@@ -100,7 +107,7 @@
 };
 
 #define CO_RETRY_TIMES  1
-#define CO_DIFF_OK      10
+#define CO_DIFF_OK      5
 #define CO_DIFF_IGNORE  23
 #define CO_MIN_LENGTH   9
 # define CO_MIN_LENGTH_BASE 5
@@ -237,28 +244,78 @@
 }
 
 /*
- * This is a reasonably severe check for a different selection of characters
- * in the old and new passwords.
+ * calculate how different two strings are in terms of the number of
+ * character removals, additions, and changes needed to go from one to
+ * the other
  */
 
-static int similar(struct cracklib_options *opt,
-		    const char *old, const char *new)
+static int distdifferent(const char *old, const char *new, int i, int j)
 {
-    int i, j;
+	char c, d;
+	if((i == 0) || (strlen(old) <= i)) {
+		c = 0;
+	} else {
+		c = old[i - 1];
+	}
+	if((j == 0) || (strlen(new) <= i)) {
+		d = 0;
+	} else {
+		d = new[j - 1];
+	}
+	return (c != d);
+}
 
-    for (i = j = 0; old[i]; i++) {
-	if (strchr (new, old[i])) {
-	    j++;
+static int distcalculate(int **distances, const char *old, const char *new, int i, int j)
+{
+	int tmp = 0;
+	if(distances[i][j] != -1)
+		return distances[i][j];
+	tmp =          distcalculate(distances, old, new, i - 1, j - 1);
+	tmp = MIN(tmp, distcalculate(distances, old, new,     i, j - 1));
+	tmp = MIN(tmp, distcalculate(distances, old, new, i - 1,     j));
+	tmp += distdifferent(old, new, i, j);
+	distances[i][j] = tmp;
+	return tmp;
+}
+
+static int distance(const char *old, const char *new)
+{
+	int **distances = NULL;
+	int m, n, i, j, r;
+
+	m = strlen(old);
+	n = strlen(new);
+	distances = malloc(sizeof(int*) * (m + 1));
+	for(i = 0; i <= m; i++) {
+		distances[i] = malloc(sizeof(int) * (n + 1));
+		for(j = 0; j <= n; j++) {
+			distances[i][j] = -1;
+		}
 	}
-    }
+	for(i = 0; i <= m; i++) {
+		distances[i][0] = i;
+	}
+	for(j = 0; j <= n; j++) {
+		distances[0][j] = j;
+	}
+	distances[0][0] = 0;
 
-    if (((i-j) >= opt->diff_ok)
-	|| (strlen(new) >= (j * 2))
-	|| (strlen(new) >= opt->diff_ignore)) {
-	/* passwords are not very similar */
-	return 0;
-    }
+	r = distcalculate(distances, old, new, m, n);
+
+	for(i = 0; i <= m; i++) {
+		free(distances[i]);
+	}
+	free(distances);
 
+	return r;
+}
+
+static int similar(struct cracklib_options *opt, const char *old, const char *new)
+{
+    if (distance(old, new) >= opt->diff_ok)
+	return 0;
+    if (strlen(new) >= (strlen(old) * 2))
+	return 0;
     /* passwords are too similar */
     return 1;
 }
@@ -475,11 +532,10 @@
     options.use_authtok = CO_USE_AUTHTOK;
     memset(options.prompt_type, 0, BUFSIZ);
     
+    strcpy(options.prompt_type,"UNIX");
     ctrl = _pam_parse(&options, argc, argv);
 
     D(("called."));
-    if (!options.prompt_type[0])
-        strcpy(options.prompt_type,"UNIX");
 
     if (flags & PAM_PRELIM_CHECK) {
         /* Check for passwd dictionary */       
